/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"
#include "cache.h"
#include "loader.h"
#include "syscall.h"
#include "bpred.h"
#include "resource.h"
#include "bitmap.h"
#include "sim.h"

/* cycle counter */
static SS_TIME_TYPE cycle = 0;

/* instruction sequence counter */
static unsigned long inst_seq = 0;

/* track number of insn and refs */
static SS_COUNTER_TYPE num_insn = 0;
static SS_COUNTER_TYPE num_total_insn = 0;
static SS_COUNTER_TYPE num_refs = 0;
static SS_COUNTER_TYPE num_loads = 0;
static SS_COUNTER_TYPE num_branches = 0;

static struct cache *icache = NULL;
static struct cache *dcache = NULL;

static struct bpred *pred = NULL;


/* exit when this becomes non-zero */
int sim_exit_now = FALSE;

/* longjmp here when simulation is completed */
jmp_buf sim_exit_buf;

/* speculation mode, non-zero when mis-speculating, i.e., executing
   instructions down the wrong path, thus recover will eventually have to
   occur that resets processor register and memory state back to the last
   precise state */
static int spec_mode = FALSE;

/* block access latency */
static int blk_access_latency = 6;

void ruu_recover(int branch_index);
void tracer_recover(void);

static unsigned long
iblk_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE baddr, int bsize,
	      struct cache_blk *blk, SS_TIME_TYPE now)
{
  mem_access(cmd, baddr, &blk->data[0], bsize);
  return (cmd == Read) ? blk_access_latency : 0;
}

static unsigned long
dblk_access_fn(enum mem_cmd cmd, SS_ADDR_TYPE baddr, int bsize,
	      struct cache_blk *blk, SS_TIME_TYPE now)
{
  /* mem_access(cmd, baddr, &blk->data[0], bsize); */
  return (cmd == Read) ? blk_access_latency : 0;
}

static int ruu_inorder_issue = FALSE;
static int ruu_perfect_prediction = FALSE;

static unsigned ruu_issue_width = 4;
static unsigned ruu_decode_width = 4;

/* cycles until fetch issues resume */
static unsigned ruu_fetch_issue_delay = 0;
static unsigned ruu_branch_penalty = 3;

static SS_COUNTER_TYPE ruu_num_rr_loads = 0;

static SS_COUNTER_TYPE ruu_num_rr_stores = 0;

static SS_COUNTER_TYPE ruu_cache_accesses = 0;
static SS_COUNTER_TYPE ruu_extra_cache_accesses = 0;

static SS_COUNTER_TYPE ruu_num_disp_loads = 0;
static SS_COUNTER_TYPE ruu_num_direct_loads = 0;
static SS_COUNTER_TYPE ruu_num_disp_stores = 0;
static SS_COUNTER_TYPE ruu_num_direct_stores = 0;

static SS_COUNTER_TYPE ruu_num_sp_loads = 0;
static SS_COUNTER_TYPE ruu_num_sp_stores = 0;

static SS_COUNTER_TYPE ruu_one_non_reg = 0;
static SS_COUNTER_TYPE ruu_one_non_rr_reg = 0;
static SS_COUNTER_TYPE ruu_one_non_store = 0;

/* inst tag type, used to tag an operation instance in the RUU */
typedef unsigned long INST_TAG_TYPE;

/* inst sequence type, used to order instructions in the ready list, if
   this rolls over the ready list order temporarily will get messed up,
   but execution will continue and complete correctly */
typedef unsigned long INST_SEQ_TYPE;


/* total input dependencies possible */
#define MAX_IDEPS		3

/* total output dependencies possible */
#define MAX_ODEPS		2

struct RUU_station {
  /* inst info */
  SS_INST_TYPE IR;			/* instruction bits */
  enum ss_opcode op;			/* instruction opcode */
  SS_ADDR_TYPE PC, next_PC, pred_PC;	/* inst PC, {predicted} next PC */
  int in_LSQ;				/* non-zero if op is in LSQ */
  int ea_comp;				/* non-zero if op is an addr comp */
  int recover_inst;			/* start of mis-speculation? */
  int spec_mode;			/* non-zero if issued in spec_mode */
  SS_ADDR_TYPE addr;			/* effective address for ld/st's */
  INST_TAG_TYPE tag;			/* RUU slot tag, increment to
					   squash operation */
  INST_SEQ_TYPE seq;			/* instruction sequence, used to
					   sort the ready list */

  /* inst status */
  int queued;				/* operands ready and queued */
  int issued;				/* operation is/was executing */
  int completed;			/* operation has completed execution */

  /* output dependency list */
  int onames[MAX_ODEPS];		/* output names (NA for no result) */
  struct RS_link *odep_list[MAX_ODEPS];	/* link to consuming op chain */

  /* input dependent links */
  int idep_ready[MAX_IDEPS];		/* input operand ready? */
};

/* non-zero if all register operands are ready, update with MAX_IDEPS */
#define OPERANDS_READY(RS)						\
  ((RS)->idep_ready[0] && (RS)->idep_ready[1] && (RS)->idep_ready[2])

/* register update unit, combination of reservation stations and reorder
   buffer device, organized as a circular queue */
static int RUU_size = 8;		/* total entries in RUU */
static struct RUU_station *RUU;		/* register update unit */
static int RUU_head, RUU_tail;		/* RUU head and tail pointers */
static int RUU_num;			/* number of entries in RUU */

/* allocate and initialize the load/store queue */
void
ruu_init(void)
{
  RUU = calloc(RUU_size, sizeof(struct RUU_station));
  if (!RUU)
    fatal("out of virtual memory");

  RUU_num = 0;
  RUU_head = RUU_tail = 0;
}


/* load/store queue: holds loads and stores in program order, indicating
   status of load/store access:

     - issued: address computation complete, memory access in progress
     - completed: memory access has completed, stored value available
     - squashed: memory access was squashed, ignore this entry

     loads may execute when:
       0) register operands are ready, and
       1) memory operands are ready (no earlier unresolved store)
     loads are serviced by:
       1) previous store at same address in LSQ (1 cycle), or
       2) data cache (1 cycle + miss latency)

     stores may execute when:
       0) register operands are ready
     stores are serviced by:
       1) depositing store value into the load/store queue
       2) writing store value to the store buffer (plus tag check) at commit
       3) writing store buffer entry to data cache when cache is free

   NOTE: the load/store queue can bypass a store value to a load in the same
     cycle the store executes (using a bypass network), thus stores complete
     in effective zero time after their effective address is known
*/
static int LSQ_size = 4;		/* total entries in LSQ */
static struct RUU_station *LSQ;		/* load/store queue */
static int LSQ_head, LSQ_tail;		/* LSQ head and tail pointers */
static int LSQ_num;			/* num of entries currently in LSQ */

/* input dependencies for stores in the LSQ:
     idep #0 - operand input (value that is store'd)
     idep #1 - effective address input (address of store operation)
*/
#define STORE_OP_INDEX			0
#define STORE_ADDR_INDEX		1

#define STORE_OP_READY(RS)		((RS)->idep_ready[STORE_OP_INDEX])
#define STORE_ADDR_READY(RS)		((RS)->idep_ready[STORE_ADDR_INDEX])

/* allocate and initialize the load/store queue */
static void
lsq_init(void)
{
  LSQ = calloc(LSQ_size, sizeof(struct RUU_station));
  if (!LSQ)
    fatal("out of virtual memory");

  LSQ_num = 0;
  LSQ_head = LSQ_tail = 0;
}


/* simulator-specific options */
char *sim_optstring = "190d:i:b:s:m:j:W:D:L:R:";

static int ruu_include_spec = TRUE;

void
sim_options(int argc, char **argv)
{
  char c;

  /* parse options */
  getopt_init();
  while ((c = getopt_next(argc, argv, sim_optstring)) != EOF)
    {
      switch (c) {
      case '0':
	ruu_include_spec = FALSE;
	info("not issuing mis-speculated instructions...");
	break;
      case '1':
	ruu_inorder_issue = TRUE;
	break;
      case '9':
	ruu_perfect_prediction = TRUE;
	info("running with perfect branch prediction...");
	break;
      case 'b':
	/* use BTB predictor */
	pred = bpred_create(BPred2bitBTB, atoi(getopt_arg));
	break;
      case 'i':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a I-cache */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad I-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
	  icache =
	    cache_create(name, nsets, 2*bsize, /* balloc */TRUE, /* usize */0,
			 assoc, cache_char2policy(c), iblk_access_fn,
			 /* hit latency */1);
	  info("** block size automatically doubled to %d bytes...", 2*bsize);
	}
	break;
      case 'd':
	{
	  char name[128], c;
	  int nsets, bsize, assoc;

	  /* use a D-cache */
	  if (sscanf(getopt_arg, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad D-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
	  dcache =
	    cache_create(name, nsets, bsize, /* balloc */FALSE, /* usize */0,
			 assoc, cache_char2policy(c), dblk_access_fn,
			 /* hit latency */1);
	}
	break;
      case 'm':
	if (icache || dcache)
	  fatal("specify `-m' before `-i' or `-d'");
	blk_access_latency = atoi(getopt_arg);
	break;
      case 'j':
	ruu_branch_penalty = atoi(getopt_arg);
	if (ruu_branch_penalty < 1)
	  fatal("mis-prediction penalty must be at least 1 cycle");
	break;
      case 'W':
	ruu_issue_width = atoi(getopt_arg);
	if (ruu_issue_width < 1
	    || (ruu_issue_width & (ruu_issue_width-1)) != 0)
	  fatal("issue width must be positive non-zero and a power of two");
	break;
      case 'D':
	ruu_decode_width = atoi(getopt_arg);
	if (ruu_decode_width < 1
	    || (ruu_decode_width & (ruu_decode_width-1)) != 0)
	  fatal("decode width must be positive non-zero and a power of two");
	break;
      case 'R':
	RUU_size = atoi(getopt_arg);
	if (RUU_size < 2 || (RUU_size & (RUU_size-1)) != 0)
	  fatal("RUU size must be positive non-zero and a power of two");
	break;
      case 'L':
	LSQ_size = atoi(getopt_arg);
	if (LSQ_size < 2 || (LSQ_size & (LSQ_size-1)) != 0)
	  fatal("LSQ size must be positive non-zero and a power of two");
	break;
      case 's':
	/* use static predictor */
	if (!strcmp(getopt_arg, "taken"))
	  pred = bpred_create(BPredTaken, 0);
	else if (!strcmp(getopt_arg, "nottaken"))
	  pred = bpred_create(BPredNotTaken, 0);
	else
	  fatal("cannot grok predictor type");
	break;
      }
    }
}

void
sim_config(FILE *stream)
{
  if (icache)
    cache_config(icache, stream);
  if (dcache)
    cache_config(dcache, stream);
  if (pred)
    bpred_config(pred, stream);
  if (ruu_inorder_issue)
    info("using in-order issue...");
#if 0
  info("issuing up to %d instructions per cycle", issue_width);
  info("cache miss latency = %d cycles", blk_access_latency);
  info("mis-prediction penalty = %d cycles", num_cycle_MP);
  if (flush_on_syscalls)
    info("flushing D-cache on system calls...");
  else
    info("running system calls through warm D-cache...");
  if (tlb)
    cache_config(tlb, stream);
  else
    info("using no predictor...");
  if (contiguous_fetch)
    info("breaking I-fetch on branches...");
  else
    info("*not* breaking I-fetch on branches...");
#endif
}

/* set to non-zero when simulator should dump statistics */
int sim_dump_stats = FALSE;

void
sim_stats(FILE *stream)
{
  extern time_t start_time;
  time_t elapsed = MAX(time((time_t *)NULL) - start_time, 1);

  info("sim: executed %.0f instructions (%.0f)",
       (double)num_insn, (double)num_total_insn);
  info("sim: executed %.0f cycles", (double)cycle);
  info("sim: IPC: %f  CPI: %f",
       (double)num_insn/(double)cycle, (double)cycle/(double)num_insn);
  info("sim: accessed (data) memory %.0f times (%.0f loads  %.0f stores)",
       (double)num_refs, (double)num_loads, (double)(num_refs-num_loads));
  info("sim: simulation time: %s (%f insts/sec)",
       elapsed_time(elapsed), (double)num_insn/(double)elapsed);
  info("sim: executed %.0f branches (%f insts/branch)",
       (double)num_branches, (double)num_insn/(double)num_branches);
  if (icache)
    cache_stats(icache, stream);
  if (dcache)
    cache_stats(dcache, stream);
  if (pred)
    bpred_stats(pred, stream);
  info("B/W: %.0f access  %.0f mis-spec  %f%% overhead",
       (double)(ruu_cache_accesses+ruu_extra_cache_accesses),
       (double)ruu_extra_cache_accesses,
       100.0 * (double)ruu_extra_cache_accesses/(double)ruu_cache_accesses);
  info("B/W no reg: %.0f access  %.0f mis-spec  %f%% overhead",
       (double)(ruu_cache_accesses+(ruu_extra_cache_accesses-ruu_one_non_reg)),
       (double)(ruu_extra_cache_accesses-ruu_one_non_reg),
       100.0 * (double)(ruu_extra_cache_accesses-ruu_one_non_reg)/(double)ruu_cache_accesses);
  info("loads: %f%% disp  %f%% direct  %f%% R+R",
       100.0 * (double)ruu_num_disp_loads/(double)num_loads,
       100.0 * (double)ruu_num_direct_loads/(double)num_loads,
       100.0 * (double)ruu_num_rr_loads/(double)num_loads);
  info("stores: %f%% disp  %f%% direct  %f%% R+R",
       100.0 * (double)ruu_num_disp_stores/(double)(num_refs-num_loads),
       100.0 * (double)ruu_num_direct_stores/(double)(num_refs-num_loads),
       100.0 * (double)ruu_num_rr_stores/(double)(num_refs-num_loads));

  info("loads: %f%% gp  %f%% sp  %f%% general",
       100.0 * (double)ruu_num_direct_loads/(double)num_loads,
       100.0 * (double)ruu_num_sp_loads/(double)num_loads,
       100.0 * (double)(num_loads-ruu_num_sp_loads-ruu_num_direct_loads)/(double)num_loads);
  info("stores: %f%% gp  %f%% sp  %f%% general",
       100.0 * (double)ruu_num_direct_stores/(double)(num_refs-num_loads),
       100.0 * (double)ruu_num_sp_stores/(double)(num_refs-num_loads),
       100.0 * (double)((num_refs-num_loads)-ruu_num_sp_stores-ruu_num_direct_stores)/(double)(num_refs-num_loads));

}


/* functional unit resource pool */
static struct res_pool *fu_pool = NULL;

/* functional unit resource configuration */
struct res_desc fu_config[] = {
  {
    "integer-ALU",
    4,
    0,
    {
      { IntALU, 1, 1 }
    }
  },
  {
    "integer-MULT/DIV",
    1,
    0,
    {
      { IntMULT, 3, 1 },
      { IntDIV, 20, 19 }
    }
  },
  {
    "memory-port",
    2,
    0,
    {
      { RdPort, 1, 1 },
      { WrPort, 1, 1 }
    }
  },
  {
    "FP-adder",
    4,
    0,
    {
      { FloatADD, 2, 1 },
      { FloatCMP, 2, 1 },
      { FloatCVT, 2, 1 }
    }
  },
  {
    "FP-MULT/DIV",
    1,
    0,
    {
      { FloatMULT, 4, 1 },
      { FloatDIV, 12, 12 },
      { FloatSQRT, 24, 24 }
    }
  },
};

/* number of free load/store pipes in the last cycle */
static int ruu_free_ldst_fu_last_cycle;

/* service all functional unit release events */
static void
ruu_release_fu(void)
{
  int i;

  ruu_free_ldst_fu_last_cycle = 0;
  for (i=0; i<fu_pool->num_resources; i++)
    {
      /* is this a free load/store pipe? */
      if (fu_pool->resources[i].x[0].class == RdPort
	  && fu_pool->resources[i].busy == 0)
	ruu_free_ldst_fu_last_cycle++;

      /* counting off an extra cache access? */
      if (fu_pool->resources[i].x[0].class == RdPort)
	{
	  if (fu_pool->resources[i].busy > 1)
	    ruu_extra_cache_accesses++;
	  if (fu_pool->resources[i].busy == 1)
	    ruu_cache_accesses++;
	}

      /* resource is released when BUSY hits zero */
      if (fu_pool->resources[i].busy > 0)
	fu_pool->resources[i].busy--;
    }
}




/* reservation station link: element of a RUU reservation station list, since
   entries in the RUU can be squashed and reused without regard for the lists
   that point to it, a link is not valid unless the link tag number matches
   the RUU reservation tag number */
struct RS_link {
  struct RS_link *next;			/* next entry in list */
  struct RUU_station *rs;		/* referenced RUU resv station */
  INST_TAG_TYPE tag;			/* inst instance sequence number */
  union {
    SS_TIME_TYPE when;			/* time stamp of entry (for eventq) */
    INST_SEQ_TYPE seq;			/* inst sequence */
    int opnum;				/* input/output operand number */
  } x;
};

/* RS link free list */
static struct RS_link *rslink_free_list;

/* total RS links allocated at program start */
#define MAX_RS_LINKS			2048

/* NULL value for an RS link */
#define RSLINK_NULL			((struct RS_link){ NULL, NULL, 0 })

/* NULL value for an RS link */
#define RSLINK_INIT(RS)		((struct RS_link){ NULL, (RS), (RS)->tag })

/* non-zero if RS link is NULL */
#define RSLINK_IS_NULL(LINK)		((LINK)->rs == NULL)

/* non-zero if RS link is to a valid (non-squashed) entry */
#define RSLINK_VALID(LINK)		((LINK)->tag == (LINK)->rs->tag)

/* extra RUU reservation station pointer */
#define RSLINK_RS(LINK)			((LINK)->rs)

/* get a new RS link record */
#define RSLINK_NEW(RS)							\
  ({ struct RS_link *n_link;						\
     if (!rslink_free_list)						\
       panic("out of RS links");					\
     n_link = rslink_free_list;						\
     rslink_free_list = rslink_free_list->next;				\
     n_link->next = NULL;						\
     n_link->rs = (RS); n_link->tag = n_link->rs->tag;			\
     n_link;								\
  })

/* free an RS link record */
#define RSLINK_FREE(LINK)						\
  ({  struct RS_link *f_link = (LINK);					\
      f_link->rs = NULL; f_link->tag = 0;				\
      f_link->next = rslink_free_list;					\
      rslink_free_list = f_link;					\
   })

/* free an RS link list */
#define RSLINK_FREE_LIST(LINK)						\
  ({  struct RS_link *fl_link, *fl_link_next;				\
      for (fl_link=(LINK); fl_link; fl_link=fl_link_next)		\
	{								\
	  fl_link_next = fl_link->next;					\
	  RSLINK_FREE(fl_link);						\
	}								\
   })

void
rslink_init(int nlinks)
{
  int i;
  struct RS_link *link;

  rslink_free_list = NULL;
  for (i=0; i<nlinks; i++)
    {
      link = calloc(1, sizeof(struct RS_link));
      if (!link)
	fatal("out of virtual memory");
      link->next = rslink_free_list;
      rslink_free_list = link;
    }
}




/* pending event queue, sorted from soonest to latest event (in time) */
static struct RS_link *event_queue;

/* initialize the event queue structures */
static void
eventq_init(void)
{
  event_queue = NULL;
}

/* insert an event for RS into the event queue, event queue is sorted from
   earliest to latest event, event and associated side-effects will be
   apparent at the start of cycle WHEN */
static void
eventq_queue_event(struct RUU_station *rs, SS_TIME_TYPE when)
{
  struct RS_link *prev, *ev, *new_ev;

  if (rs->completed)
    panic("event completed");

  if (when <= cycle)
    panic("event occurred in the past");

  /* get a free event record */
  new_ev = RSLINK_NEW(rs);
  new_ev->x.when = when;

  /* locate insertion point */
  for (prev=NULL, ev=event_queue;
       ev && ev->x.when < when;
       prev=ev, ev=ev->next);

  if (prev)
    {
      /* insert middle or end */
      new_ev->next = prev->next;
      prev->next = new_ev;
    }
  else
    {
      /* insert at beginning */
      new_ev->next = event_queue;
      event_queue = new_ev;
    }
}

/* return the next event that has already occurred, returns NULL when all
   remaining events are in the future */
static struct RUU_station *
eventq_next_event(void)
{
  struct RS_link *ev;

  if (event_queue && event_queue->x.when <= cycle)
    {
      /* unlink and return first event on priority list */
      ev = event_queue;
      event_queue = event_queue->next;

      /* event still valid? */
      if (RSLINK_VALID(ev))
	{
	  struct RUU_station *rs = RSLINK_RS(ev);

	  /* reclaim event record */
	  RSLINK_FREE(ev);

	  /* event is valid, return resv station */
	  return rs;
	}
      else
	{
	  /* reclaim event record */
	  RSLINK_FREE(ev);

	  /* receiving inst was squashed, return next event */
	  return eventq_next_event();
	}
    }
  else
    {
      /* no event or no event is ready */
      return NULL;
    }
}




/* RS ready queue: all entries have register and memory input dependencies
   satisfied, entries may be issued whenever resources are available, entries
   may get squashed while in this list, ready queue is fully constructed
   each cycle before any operation is issued from it -- this ensures that
   instruction issue priorities are observed properly
*/
static struct RS_link *ready_queue;

/* initialize the event queue structures */
static void
readyq_init(void)
{
  ready_queue = NULL;
}

/* insert ready node into the ready list */
static void
readyq_enqueue(struct RUU_station *rs)
{
  struct RS_link *prev, *node, *new_node;

  /* node is now queued */
  if (rs->queued)
    panic("node is already queued");
  rs->queued = TRUE;

  /* get a free ready list node */
  new_node = RSLINK_NEW(rs);
  new_node->x.seq = rs->seq;

  /* locate insertion point */
  if (rs->in_LSQ || SS_OP_FLAGS(rs->op) & (F_LONGLAT|F_CTRL))
    {
      /* insert loads/stores and long latency ops at the head of the queue */
      prev = NULL;
      node = ready_queue;
    }
  else
    {
      /* otherwise insert in program order (earliest seq first) */
      for (prev=NULL, node=ready_queue;
	   node && node->x.seq < rs->seq;
	   prev=node, node=node->next);
    }

  if (prev)
    {
      /* insert middle or end */
      new_node->next = prev->next;
      prev->next = new_node;
    }
  else
    {
      /* insert at beginning */
      new_node->next = ready_queue;
      ready_queue = new_node;
    }
}


static void
lsq_refresh(void)
{
  int i, index;

  /* scan entire queue for ready loads: scan from oldest instruction
     (head) until we reach the tail or an unresolved store, after which no
     other instruction will become ready */
  for (i=0, index=LSQ_head;
       i < LSQ_num;
       i++, index=(index + 1) % LSQ_size)
    {
      /* ignore entry? */
      if (/* completed? */LSQ[index].completed)
	continue;

      /* terminate search for ready loads after first unresolved store,
         as no later load could be resolved in its presence */
      if (/* store? */
	  ((SS_OP_FLAGS(LSQ[index].op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
	  && /* unresolved */!STORE_ADDR_READY(&LSQ[index]))
	break;

      if (/* load? */
	  ((SS_OP_FLAGS(LSQ[index].op) & (F_MEM|F_LOAD)) == (F_MEM|F_LOAD))
	  && /* queued? */!LSQ[index].queued
	  && /* waiting? */!LSQ[index].issued
	  && /* regs ready? */OPERANDS_READY(&LSQ[index]))
	{
	  /* no conflict, put the load on the ready queue */
	  readyq_enqueue(&LSQ[index]);
	}
    }
}




/* attempt to issue all operations in the ready list */
static void
ruu_issue(void)
{
  int i, load_lat;
  struct RS_link *node, *next_node;
  struct res_template *fu;

  /* FIXME: could be a little more efficient when scanning the ready queue */

  /* copy and then blow away the ready list */
  node = ready_queue;
  ready_queue = NULL;

  for (; node; node = next_node)
    {
      next_node = node->next;

      /* still valid? */
      if (RSLINK_VALID(node))
	{
	  struct RUU_station *rs = RSLINK_RS(node);

	  /* issue operation, both reg and mem deps have been satisfied */
	  if (!OPERANDS_READY(rs) || !rs->queued
	      || rs->issued || rs->completed)
	    panic("issued inst !ready, issued, or completed");

	  /* node is now un-queued */
	  rs->queued = FALSE;

	  if (rs->in_LSQ
	      && ((SS_OP_FLAGS(rs->op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE)))
	    {
	      /* stores complete in effectively zero time, result is
		 written into the load/store queue */
	      rs->issued = TRUE;
	      rs->completed = TRUE;

	      if (rs->onames[0] || rs->onames[1])
		panic("store creates result");

	      if (rs->recover_inst)
		panic("mis-predicted store");
	    }
	  else
	    {
	      /* issue the instruction to a functional unit */
	      if (SS_OP_FUCLASS(rs->op) != NA)
		{
		  fu = res_get(fu_pool, SS_OP_FUCLASS(rs->op));
		  if (fu)
		    {
		      /* issue inst to functional unit */
		      rs->issued = TRUE;

		      /* reserve the functional unit */
		      if (fu->master->busy)
			panic("functional unit already in use");

		      /* schedule functional unit release event */
		      fu->master->busy = fu->issuelat;

		      /* schedule a result event */
		      if (rs->in_LSQ
			  && ((SS_OP_FLAGS(rs->op) & (F_MEM|F_LOAD))
			      == (F_MEM|F_LOAD)))
			{
			  /* determine cache access latency: scan LSQ */
			  load_lat = 0;
			  i = (rs - LSQ);
			  if (i != LSQ_head)
			    {
			      for (;;)
				{
				  /* go to next earlier LSQ entry */
				  i = (i + (LSQ_size-1)) % LSQ_size;

				  if ((SS_OP_FLAGS(LSQ[i].op) & F_STORE)
				      && (LSQ[i].addr == rs->addr))
				    {
				      /* hit in the LSQ */
				      load_lat = 1;
				      break;
				    }

				  /* scan finished? */
				  if (i == LSQ_head)
				    break;
				}
			    }

			  if (!load_lat)
			    {
			      /* go to the data cache */
			      if (dcache
				  && (rs->addr >= ld_data_base
				      && rs->addr < ld_stack_base))
				{
				  /* access the cache if non-faulting */
				  load_lat =
				    cache_access(dcache, Read,
						 (rs->addr & ~3), NULL, 4,
						 cycle, NULL, NULL);
				}
			      else
				load_lat = fu->oplat;
			    }

			  /* use deterministic functional unit latency */
			  eventq_queue_event(rs, cycle + load_lat);
			}
		      else
			{
			  /* use deterministic functional unit latency */
			  eventq_queue_event(rs, cycle + fu->oplat);
			}
		    }
		  else
		    {
		      /* insufficient functional unit resources, put op back
			 onto the ready list */
		      readyq_enqueue(rs);
		    }
		}
	      else
		{
		  /* FIXME: need better solution for these */
		  /* issue inst to functional unit */
		  rs->issued = TRUE;

		  /* schedule a result event */
		  eventq_queue_event(rs, cycle + 1);
		}
	    }
	}
      /* else, RUU entry was squashed */

      /* reclaim ready list entry */
      RSLINK_FREE(node);
    }
}




/* create vector, maps a logical register to a creator in the RUU (and
   specific output operand) or the architected register file (if RS_link
   is NULL) */
struct CV_link {
  struct RUU_station *rs;		/* creator's reservation station */
  int odep_num;				/* specific output operand */
};
#define CVLINK_NULL		((struct CV_link){ NULL, 0 })
#define CVLINK_NEW(RS,ONUM)	((struct CV_link){ (RS), (ONUM) })
#define CV_BMAP_SZ		(BITMAP_SIZE(SS_TOTAL_REGS))
static BITMAP_TYPE(SS_TOTAL_REGS, use_spec_cv);
static struct CV_link create_vector[SS_TOTAL_REGS];
static SS_TIME_TYPE create_vector_rt[SS_TOTAL_REGS];
static struct CV_link spec_create_vector[SS_TOTAL_REGS];
static SS_TIME_TYPE spec_create_vector_rt[SS_TOTAL_REGS];

#define CREATE_VECTOR(N)	(BITMAP_SET_P(use_spec_cv, CV_BMAP_SZ, (N))\
				 ? spec_create_vector[N]		\
				 : create_vector[N])
#define CREATE_VECTOR_RT(N)	(BITMAP_SET_P(use_spec_cv, CV_BMAP_SZ, (N))\
				 ? spec_create_vector_rt[N]		\
				 : create_vector_rt[N])
#define SET_CREATE_VECTOR(N, L)	(spec_mode				\
				 ? (BITMAP_SET(use_spec_cv, CV_BMAP_SZ, (N)),\
				    spec_create_vector[N] = (L))	\
				 : (create_vector[N] = (L)))

void
cv_init(void)
{
  int i;

  for (i=0; i<SS_TOTAL_REGS; i++)
    {
      create_vector[i] = CVLINK_NULL;
      create_vector_rt[i] = 0;
      spec_create_vector[i] = CVLINK_NULL;
      spec_create_vector_rt[i] = 0;
    }
  BITMAP_CLEAR_MAP(use_spec_cv, CV_BMAP_SZ);
}

/* writeback completed ops from the functional units to RUU */
static void
ruu_writeback(void)
{
  int i;
  struct RUU_station *rs;

  /* service completed events */
  while ((rs = eventq_next_event()))
    {
      /* RS has completed execution and (possibly) produced a result */
      if (!OPERANDS_READY(rs) || rs->queued || !rs->issued || rs->completed)
	panic("inst completed and !ready, !issued, or completed");

      /* non-speculative result */
      rs->completed = TRUE;

      /* mis-predicted branch? */
      if (rs->recover_inst)
	{
	  if (rs->in_LSQ)
	    panic("mis-predicted load or store?!?!?");

	  /* recover processor state and reinit fetch to correct path */
	  ruu_recover(rs - RUU);
	  tracer_recover();

	  /* stall fetch until I-fetch and I-decode recover */
	  ruu_fetch_issue_delay = ruu_branch_penalty;

	  /* continue writeback of the branch/control instruction */
	}

      /* broadcast results to consuming operations */
      for (i=0; i<MAX_ODEPS; i++)
	{
	  if (rs->onames[i] != NA)
	    {
	      struct CV_link link;
	      struct RS_link *olink, *olink_next;

	      if (rs->spec_mode)
		{
		  /* update the create vector, future operations get value from
		     later creator or architected register file */
		  link = spec_create_vector[rs->onames[i]];
		  if (/* !NULL */link.rs
		      && /* refs RS */(link.rs == rs && link.odep_num == i))
		    {
		      spec_create_vector[rs->onames[i]] = CVLINK_NULL;
		      spec_create_vector_rt[rs->onames[i]] = cycle;
		    }
		  /* else, creator invalidated or there is another creator */
		}
	      else
		{
		  /* update the create vector, future operations get value from
		     later creator or architected register file */
		  link = create_vector[rs->onames[i]];
		  if (/* !NULL */link.rs
		      && /* refs RS */(link.rs == rs && link.odep_num == i))
		    {
		      create_vector[rs->onames[i]] = CVLINK_NULL;
		      create_vector_rt[rs->onames[i]] = cycle;
		    }
		  /* else, creator invalidated or there is another creator */
		}

	      /* walk output list, queue up ready operations */
	      for (olink=rs->odep_list[i]; olink; olink=olink_next)
		{
		  if (RSLINK_VALID(olink))
		    {
		      if (olink->rs->idep_ready[olink->x.opnum])
			panic("output dependence already satisfied");

		      /* input is now ready */
		      olink->rs->idep_ready[olink->x.opnum] = TRUE;

		      if (OPERANDS_READY(olink->rs))
			{
			  if (!olink->rs->in_LSQ
			      || ((SS_OP_FLAGS(olink->rs->op)&(F_MEM|F_STORE))
				  == (F_MEM|F_STORE)))
			    readyq_enqueue(olink->rs);
			  /* else, ld op, issued when no mem conflict */
			}
		    }

		  /* grab link to next element prior to free */
		  olink_next = olink->next;

		  /* free dependence link element */
		  RSLINK_FREE(olink);
		}
	      /* blow away the consuming op list */
	      rs->odep_list[i] = NULL;
	    }
	}
   }
}




/* commit entries from the RUU to the architected reg file */
static void
ruu_commit(void)
{
  /* all values must be retired to the architected reg file in program order */
  while (RUU_num > 0)
    {
      if (!RUU[RUU_head].completed)
	{
	  /* at least RUU entry must be complete */
	  break;
	}

      /* load/stores must retire load/store queue entry as well */
      if (RUU[RUU_head].ea_comp)
	{
	  /* load/store, retire head of LSQ as well */
	  if (LSQ_num<=0 || !LSQ[LSQ_head].in_LSQ)
	    panic("RUU out of sync with LSQ");

	  /* load/store operation must be complete */
	  if (!LSQ[LSQ_head].completed)
	    {
	      /* load/store operation is not yet complete */
	      break;
	    }

	  if ((SS_OP_FLAGS(LSQ[LSQ_head].op) & (F_MEM|F_STORE))
	      == (F_MEM|F_STORE))
	    {
	      struct res_template *fu;

	      /* try to get a store port (functional unit allocation) */
	      fu = res_get(fu_pool, SS_OP_FUCLASS(LSQ[LSQ_head].op));
	      if (fu)
		{
		  /* reserve the functional unit */
		  if (fu->master->busy)
		    panic("functional unit already in use");

		  /* schedule functional unit release event */
		  fu->master->busy = fu->issuelat;

		  /* go to the data cache */
		  if (dcache)
		    {
		      /* commit store value to D-cache */
		      cache_access(dcache, Write, (LSQ[LSQ_head].addr & ~3),
				   NULL, 4, cycle, NULL, NULL);
		    }
		}
	      else
		{
		  /* no store ports left */
		  break;
		}
	    }

	  /* invalidate load/store operation instance */
	  LSQ[LSQ_head].tag++;

	  /* commit head of LSQ as well */
	  LSQ_head = (LSQ_head + 1) % LSQ_size;
	  LSQ_num--;
	}

      /* invalidate RUU operation instance */
      RUU[RUU_head].tag++;

      /* commit head entry of RUU */
      RUU_head = (RUU_head + 1) % RUU_size;
      RUU_num--;
    }
}



/* recover processor state to just after completion of branch at BRANCH_INDEX
   in the RUU */
void
ruu_recover(int branch_index)
{
  int i, RUU_index = RUU_tail, LSQ_index = LSQ_tail;
  int RUU_prev_tail = RUU_tail, LSQ_prev_tail = LSQ_tail;

  /* recover from tail towards head until the branch index is reached,
     this direction ensures that the LSQ remains synchronized with
     the RUU */

  /* go to first element to squash */
  RUU_index = (RUU_index + (RUU_size-1)) % RUU_size;
  LSQ_index = (LSQ_index + (LSQ_size-1)) % LSQ_size;

  while (RUU_index != branch_index)
    {
      if (!RUU_num)
	panic("empty RUU");

      if (RUU_index == RUU_head)
	panic("RUU head and tail broken");

      if (RUU[RUU_index].ea_comp)
	{
	  if (!LSQ_num)
	    panic("RUU and LSQ out of sync");

	  /* recover any resources */
	  for (i=0; i<MAX_ODEPS; i++)
	    {
	      RSLINK_FREE_LIST(LSQ[LSQ_index].odep_list[i]);
	      /* blow away the consuming op list */
	      LSQ[LSQ_index].odep_list[i] = NULL;
	    }
      
	  /* squash this entry */
	  LSQ[LSQ_index].tag++;

	  /* go to next earlier slot */
	  LSQ_prev_tail = LSQ_index;
	  LSQ_index = (LSQ_index + (LSQ_size-1)) % LSQ_size;
	  LSQ_num--;
	}

      /* recover any resources */
      for (i=0; i<MAX_ODEPS; i++)
	{
	  RSLINK_FREE_LIST(RUU[RUU_index].odep_list[i]);
	  /* blow away the consuming op list */
	  RUU[RUU_index].odep_list[i] = NULL;
	}
      
      /* squash this entry */
      RUU[RUU_index].tag++;

      /* go to next earlier slot */
      RUU_prev_tail = RUU_index;
      RUU_index = (RUU_index + (RUU_size-1)) % RUU_size;
      RUU_num--;
    }

  /* reset head/tail pointers */
  RUU_tail = RUU_prev_tail;
  LSQ_tail = LSQ_prev_tail;

  /* revert back to last precise create vector state */
  BITMAP_CLEAR_MAP(use_spec_cv, CV_BMAP_SZ);

  /* FIXME: could reset functional units */
}



/* processor register state, precise and speculative */
#define R_BMAP_SZ	(BITMAP_SIZE(SS_NUM_REGS))
static BITMAP_TYPE(SS_NUM_REGS, use_spec_R);
static SS_WORD_TYPE spec_regs_R[SS_NUM_REGS];

#define F_BMAP_SZ	(BITMAP_SIZE(SS_NUM_REGS))
static BITMAP_TYPE(SS_NUM_REGS, use_spec_F);
static union regs_FP spec_regs_F;

static int use_spec_HI;
static SS_WORD_TYPE spec_regs_HI;

static int use_spec_LO;
static SS_WORD_TYPE spec_regs_LO;

static int use_spec_FCC;
static int spec_regs_FCC;

static SS_ADDR_TYPE pred_PC;
static SS_ADDR_TYPE recover_PC;

static SS_ADDR_TYPE fetch_regs_PC;
static SS_ADDR_TYPE fetch_pred_PC;

/* processor memory state, precise and speculative */

/* speculative memory table, access through this table when accessing memory
   in speculative mode, flush the table when recovering from mispredicted
   branches */
struct spec_mem_ent {
  struct spec_mem_ent *next;
  SS_ADDR_TYPE addr;
  unsigned long data;
};

/* hash size must be a power-of-two value */
#define STORE_HASH_SIZE		32
static struct spec_mem_ent *store_htable[STORE_HASH_SIZE];

static struct spec_mem_ent *bucket_free_list = NULL;

#define HASH_ADDR(ADDR)							\
  ((((ADDR) >> 24)^((ADDR) >> 16)^((ADDR) >> 8)^(ADDR)) & (STORE_HASH_SIZE-1))

static void
spec_mem_access(enum mem_cmd cmd, SS_ADDR_TYPE addr, void *p, int nbytes)
{
  int index;
  struct spec_mem_ent *ent;

/*  return;*/

  /* check alignments, even speculative this test should always pass */
  if ((nbytes & (nbytes-1)) != 0 || (addr & (nbytes-1)) != 0)
    {
      /* no can do */
      return;
    }

  /* check permissions */
  if (!((addr >= ld_text_base && addr < (ld_text_base+ld_text_size)
	 && cmd == Read)
	|| (addr >= ld_data_base && addr < ld_stack_base)))
    {
      /* no can do */
      return;
    }

  index = HASH_ADDR(addr);
  for (ent=store_htable[index]; ent; ent=ent->next)
    {
      if (ent->addr == addr)
	break;
    }

  if (!ent && cmd == Write)
    {
      if (!bucket_free_list)
	{
	  bucket_free_list = calloc(1, sizeof(struct spec_mem_ent));
	  if (!bucket_free_list)
	    fatal("out of virtual memory");
	}
      ent = bucket_free_list;
      bucket_free_list = bucket_free_list->next;
    }

  switch (nbytes) {
  case 1:
    if (cmd == Read)
      {
	if (ent)
	  *((unsigned char *)p) = *((unsigned char *)(&ent->data));
	else
	  *((unsigned char *)p) = MEM_BYTE(addr);
      }
    else
      *((unsigned char *)(&ent->data)) = *((unsigned char *)p);
    break;
  case 2:
    if (cmd == Read)
      {
	if (ent)
	  *((unsigned short *)p) = *((unsigned short *)(&ent->data));
	else
	  *((unsigned short *)p) = MEM_HALF(addr);
      }
    else
      *((unsigned short *)&ent->data) = *((unsigned short *)p);
    break;
  case 4:
    if (cmd == Read)
      {
	if (ent)
	  *((unsigned long *)p) = *((unsigned long *)&ent->data);
	else
	  *((unsigned long *)p) = MEM_WORD(addr);
      }
    else
      *((unsigned long *)&ent->data) = *((unsigned long *)p);
    break;
  default:
    panic("large accesses not supported in mis-speculative mode");
  }
}



/* recover tracer state to last precise state */
void
tracer_recover(void)
{
  int i;
  struct spec_mem_ent *ent, *ent_next;
  static void fetch_squash(void);

  if (!spec_mode)
    panic("cannot recover unless in speculative mode");
  spec_mode = FALSE;

  /* reset registers back to precise state */
  BITMAP_CLEAR_MAP(use_spec_R, R_BMAP_SZ);
  BITMAP_CLEAR_MAP(use_spec_F, F_BMAP_SZ);
  use_spec_HI = FALSE;
  use_spec_LO = FALSE;
  use_spec_FCC = FALSE;

  /* reset memory back to precise state */
  for (i=0; i<STORE_HASH_SIZE; i++)
    {
      for (ent=store_htable[i]; ent; ent=ent_next)
	{
	  ent_next = ent->next;
	  ent->next = bucket_free_list;
	  bucket_free_list = ent;
	}
    }

  /* reset PC */
  fetch_pred_PC = fetch_regs_PC = recover_PC;

  /* restart fetch */
  fetch_squash();
}

static void
tracer_init(void)
{
  int i;

  /* initially in precise mode */
  spec_mode = FALSE;
  BITMAP_CLEAR_MAP(use_spec_R, R_BMAP_SZ);
  BITMAP_CLEAR_MAP(use_spec_F, F_BMAP_SZ);
  use_spec_HI = FALSE;
  use_spec_LO = FALSE;
  use_spec_FCC = FALSE;

  /* reset memory back to precise state */
  for (i=0; i<STORE_HASH_SIZE; i++)
    store_htable[i] = NULL;
}

static inline void
ruu_link_idep(struct RUU_station *rs, int idep_num, int idep_name)
{
  struct CV_link head;
  struct RS_link *link;

  if (idep_name == NA)
    {
      /* no input dependence for this input slot */
      rs->idep_ready[idep_num] = TRUE;
      return;
    }

  /* locate creator of operand */
  head = CREATE_VECTOR(idep_name);
  if (!head.rs)
    {
      /* no active creator, use value available in architected reg file */
      rs->idep_ready[idep_num] = TRUE;
      return;
    }

  /* indicate value will be created sometime in the future */
  rs->idep_ready[idep_num] = FALSE;

  /* link onto creator's output list of dependant operand */
  link = RSLINK_NEW(rs); link->x.opnum = idep_num;
  link->next = head.rs->odep_list[head.odep_num];
  head.rs->odep_list[head.odep_num] = link;
}

static inline void
ruu_install_odep(struct RUU_station *rs, int odep_num, int odep_name)
{
  if (odep_name == NA)
    {
      /* no value created */
      rs->onames[odep_num] = NA;
      return;
    }

  /* record output name, used to update create vector at completion */
  rs->onames[odep_num] = odep_name;

  /* initialize output chain to empty list */
  rs->odep_list[odep_num] = NULL;

  /* indicate this operation is latest creator of ODEP_NAME */
  SET_CREATE_VECTOR(odep_name, CVLINK_NEW(rs, odep_num));
}


/* dependence decoders */
#define DEP_GPR(N)		(N)
#define DEP_CGPR(N)		(SS_COMP_OP != SS_COMP_NOP ? (N) : 0)
#define DEP_GPR_D(N)		((N)&~1)
#define DEP_FPR_L(N)		(((N)+32)&~1)
#define DEP_FPR_F(N)		(((N)+32)&~1)
#define DEP_FPR_D(N)		(((N)+32)&~1)
#define DEP_HI			(0+32+32)
#define DEP_LO			(1+32+32)
#define DEP_FCC			(2+32+32)
#define DEP_TMP			(3+32+32)
#define DEP_NA			(0)

/* register file accessors */
#define SET_NPC(EXPR)		(next_PC = (EXPR))
#define CPC			(regs_PC)
#define SET_CPC(EXPR)		(regs_PC = (EXPR))
#define GPR(N)			(BITMAP_SET_P(use_spec_R, R_BMAP_SZ, (N))\
				 ? spec_regs_R[N]			\
				 : regs_R[N])
#define SET_GPR(N,EXPR)		(spec_mode				\
				 ? (BITMAP_SET(use_spec_R, R_BMAP_SZ, (N)),\
				    spec_regs_R[N] = (EXPR))		\
				 : (regs_R[N] = (EXPR)))

#define FPR_L(N)		(BITMAP_SET_P(use_spec_F, F_BMAP_SZ, ((N)&~1))\
				 ? spec_regs_F.l[(N)]			\
				 : regs_F.l[(N)])
#define SET_FPR_L(N,EXPR)	(spec_mode				\
				 ? (BITMAP_SET(use_spec_F,F_BMAP_SZ,((N)&~1)),\
				    spec_regs_F.l[(N)] = (EXPR))	\
				 : (regs_F.l[(N)] = (EXPR)))
#define FPR_F(N)		(BITMAP_SET_P(use_spec_F, F_BMAP_SZ, ((N)&~1))\
				 ? spec_regs_F.f[(N)]			\
				 : regs_F.f[(N)])
#define SET_FPR_F(N,EXPR)	(spec_mode				\
				 ? (BITMAP_SET(use_spec_F,F_BMAP_SZ,((N)&~1)),\
				    spec_regs_F.f[(N)] = (EXPR))	\
				 : (regs_F.f[(N)] = (EXPR)))
#define FPR_D(N)		(BITMAP_SET_P(use_spec_F, F_BMAP_SZ, ((N)&~1))\
				 ? spec_regs_F.d[(N) >> 1]		\
				 : regs_F.d[(N) >> 1])
#define SET_FPR_D(N,EXPR)	(spec_mode				\
				 ? (BITMAP_SET(use_spec_F,F_BMAP_SZ,((N)&~1)),\
				    spec_regs_F.d[(N) >> 1] = (EXPR))	\
				 : (regs_F.d[(N) >> 1] = (EXPR)))
#define HI			(use_spec_HI ? spec_regs_HI : regs_HI)
#define SET_HI(EXPR)		(spec_mode				\
				 ? (use_spec_HI=TRUE,			\
				    spec_regs_HI = (EXPR))		\
				 : (regs_HI = (EXPR)))
#define LO			(use_spec_LO ? spec_regs_LO : regs_LO)
#define SET_LO(EXPR)		(spec_mode				\
				 ? (use_spec_LO=TRUE,			\
				    spec_regs_LO = (EXPR))		\
				 : (regs_LO = (EXPR)))
#define FCC			(use_spec_FCC ? spec_regs_FCC : regs_FCC)
#define SET_FCC(EXPR)		(spec_mode				\
				 ? (use_spec_FCC=TRUE,			\
				    spec_regs_FCC = (EXPR))		\
				 : (regs_FCC = (EXPR)))

/* storage accessors */
#define __READ_CACHE(DST_T, SRC_T, SRC)					\
  ({ SRC_T temp; num_refs++; num_loads++;				\
     if (spec_mode)							\
       spec_mem_access(Read, (addr = (SRC)), &temp, sizeof(SRC_T));	\
     else								\
       mem_access(Read, (addr = (SRC)), &temp, sizeof(SRC_T));		\
     ((unsigned long)((DST_T)temp)); })

#define READ_WORD(SRC)							\
  __READ_CACHE(unsigned long, unsigned long, (SRC))
#define READ_UNSIGNED_HALF(SRC)						\
  __READ_CACHE(unsigned long, unsigned short, (SRC))
#define READ_SIGNED_HALF(SRC)						\
  __READ_CACHE(signed long, signed short, (SRC))
#define READ_UNSIGNED_BYTE(SRC)						\
  __READ_CACHE(unsigned long, unsigned char, (SRC))
#define READ_SIGNED_BYTE(SRC)						\
  __READ_CACHE(signed long, signed char, (SRC))

#define __WRITE_CACHE(SRC, DST_T, DST)					\
  ({ DST_T temp = (DST_T)((unsigned long)(SRC));			\
     num_refs++;							\
     if (spec_mode)							\
       spec_mem_access(Write, (addr = (DST)), &temp, sizeof(DST_T));	\
     else								\
       mem_access(Write, (addr = (DST)), &temp, sizeof(DST_T)); })

#define WRITE_WORD(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned long, (DST))
#define WRITE_HALF(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned short, (DST))
#define WRITE_BYTE(SRC, DST)						\
  __WRITE_CACHE((SRC), unsigned char, (DST))

/* syscall interface */
#define SYSCALL(INST)							\
  ({ /* only execute system calls in non-speculative mode */		\
     if (spec_mode)							\
       panic("speculative syscall");					\
     ss_syscall(mem_access, INST);					\
     /* else, ignore the system call */ });

/* non-expression inst implementations */
#define DEFFU(FU,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFFU
#undef DEFINST
#undef DEFLINK
#undef CONNECT
#undef IMPL

/* a valid address? */
#define VALID_ADDR(ADDR) (SS_DATA_BASE <= (ADDR) && (ADDR) <= SS_STACK_BASE)

/* for in-order issue */
struct RS_link last_op = RSLINK_NULL;

unsigned long first_seq = 0;

static inline int
ruu_can_issue_rr(int in1, int in2, int in3)
{
  int reg_result = TRUE;

  /* conflict with register inputs? */
  if ((in1 != NA
       && (CREATE_VECTOR(in1).rs || CREATE_VECTOR_RT(in1) == cycle))
      || (in2 != NA
	  && (CREATE_VECTOR(in2).rs || CREATE_VECTOR_RT(in2) == cycle))
      || (in3 != NA
	  && (CREATE_VECTOR(in3).rs || CREATE_VECTOR_RT(in3) == cycle)))
    {
      ruu_one_non_rr_reg++;
      reg_result = FALSE;
    }

  return reg_result;
}

/* ATTN: mega-kludge */
static inline int
ruu_can_issue(SS_ADDR_TYPE addr, int in1, int in2, int in3)
{
  int i, index, reg_result = TRUE, mem_result = TRUE;

  /* conflict with register inputs? */
  if ((in1 != NA
       && (CREATE_VECTOR(in1).rs || CREATE_VECTOR_RT(in1) == cycle))
      || (in2 != NA
	  && (CREATE_VECTOR(in2).rs || CREATE_VECTOR_RT(in2) == cycle))
      || (in3 != NA
	  && (CREATE_VECTOR(in3).rs || CREATE_VECTOR_RT(in3) == cycle)))
    {
      ruu_one_non_reg++;
      reg_result = FALSE;
    }

  /* store conflict? */
  for (i=0, index=LSQ_head;
       i < LSQ_num;
       i++, index=(index + 1) % LSQ_size)
    {
      if (((SS_OP_FLAGS(LSQ[index].op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
	  && (/* issued in the same cycle */LSQ[index].seq >= first_seq
	      || /* store to unknown address */!STORE_ADDR_READY(&LSQ[index])))
	{
	  ruu_one_non_store++;
	  mem_result = FALSE;
	  break;
	}
    }

  return reg_result && mem_result;
}

/* ATTN: mega-kludge */
extern char *mru_udata;

#define DEC_BMAP_SZ	BITMAP_SIZE(SS_TOTAL_REGS)
static BITMAP_TYPE(SS_TOTAL_REGS, decode_stage_results);

static struct fetch_rec {
  SS_INST_TYPE IR;                      /* inst register */
  SS_ADDR_TYPE regs_PC, pred_PC;        /* current PC, predicted next PC */
} *fetch_data;
static int fetch_num;                 /* num entries in idecode queue */
static int fetch_tail, fetch_head;  /* head and tail pointer of queue */

static void
fetch_squash(void)
{
  fetch_num = 0;
  fetch_tail = fetch_head = 0;
}

static void
fetch_init(void)
{
  fetch_data =
    (struct fetch_rec *)calloc(ruu_decode_width, sizeof(struct fetch_rec));
  if (!fetch_data)
    fatal("out of virtual memory");
  fetch_squash();
}

void
ruu_dispatch(void)
{
  int i;
  SS_INST_TYPE inst;			/* actual instruction bits */
  enum ss_opcode op;			/* decoded opcode enum */
  int out1, out2, in1, in2, in3;	/* output/input register names */
  SS_ADDR_TYPE next_PC;			/* actual next PC address */
  SS_ADDR_TYPE addr;			/* effective address, if load/store */
  SS_ADDR_TYPE base_addr;		/* base register value */
  signed long offset;			/* offset value */
  int base_reg;				/* base register number */
  int index_reg;			/* index register number */
  struct RUU_station *rs;		/* RUU station being allocated */
  struct RUU_station *lsq;		/*   tandem LSQ station for ld/st's */
  int no_ea_dep;			/* no effective address dep? */
  int n_dispatched;			/* total inst dispatched */

  BITMAP_CLEAR_MAP(decode_stage_results, BITMAP_SIZE(SS_TOTAL_REGS));
  first_seq = inst_seq+1;

  for (i=0, n_dispatched=0;
       i < ruu_decode_width
       && n_dispatched < ruu_issue_width
       && RUU_num < RUU_size && LSQ_num < LSQ_size
       && fetch_num
       && (ruu_include_spec || !spec_mode);
       i++)
    {
      /* block until last op issues if inorder issue */
      if (ruu_inorder_issue
	  && last_op.rs && RSLINK_VALID(&last_op)
	  && !OPERANDS_READY(last_op.rs))
	{
	  /* stall until last operation is ready to issue */
	  return;
	}
	
      inst = fetch_data[fetch_head].IR;
      regs_PC = fetch_data[fetch_head].regs_PC;
      pred_PC = fetch_data[fetch_head].pred_PC;

      /* decode the inst */
      op = SS_OPCODE(inst);

      /* compute default next_PC */
      next_PC = regs_PC + sizeof(SS_INST_TYPE);

      /* drain RUU for TRAPs */
      if (SS_OP_FLAGS(op) & F_TRAP)
	{
	  if (RUU_num != 0)
	    break;

	  /* else, should not be in (mis-)speculative mode */
	  if (spec_mode)
	    panic("drained and speculative");
	}

      /* default effective address (none) */
      addr = 0;

      /* maintain $r0 semantics */
      regs_R[0] = 0; spec_regs_R[0] = 0;

      /* decode load/store details */
      base_addr = offset = base_reg = index_reg = 0;
      if (SS_OP_FLAGS(op) & F_MEM)
	{
	  if (SS_OP_FLAGS(op) & F_RR)
	    {
	      if (SS_OP_FLAGS(op) & F_LOAD)
		ruu_num_rr_loads++;
	      else
		ruu_num_rr_stores++;
	      /* R + R */
	      if (VALID_ADDR(GPR(BS)))
		{
		  base_reg = (BS);
		  index_reg = (RD);
		  base_addr = GPR(BS);
		  offset = GPR(RD);
		}
	      else if (VALID_ADDR(GPR(RD)))
		{
		  base_reg = (RD);
		  index_reg = (BS);
		  base_addr = GPR(RD);
		  offset = GPR(BS);
		}
	      else
		{
		  /* can get these for speculative accesses */
		  base_reg = 0;
		  index_reg = 0;
		  base_addr = 0;
		  offset = 0;
		}
	    }
	  else if (SS_OP_FLAGS(op) & F_DISP)
	    {
	      /* R + C */
	      base_reg = (BS);
	      index_reg = 0;
	      base_addr = GPR(BS);
	      offset = (IMM);
	      if (SS_OP_FLAGS(op) & F_LOAD)
		{
		  ruu_num_disp_loads++;
		  if (base_reg == 29)
		    ruu_num_sp_loads++;
		}
	      else
		{
		  ruu_num_disp_stores++;
		  if (base_reg == 29)
		    ruu_num_sp_stores++;
		}
	    }
	  else if (SS_OP_FLAGS(op) & F_DIRECT)
	    {
	      if (SS_OP_FLAGS(op) & F_LOAD)
		ruu_num_direct_loads++;
	      else
		ruu_num_direct_stores++;
	      /* C */
	      base_reg = 28;
	      index_reg = 0;
	      base_addr = GPR(BS);
	      offset = (IMM);
	    }
	  else
	    panic("cannot decode load/store operation");
	}

      /* more decoding and execution */
      switch (op) {
/* avoid divides by zero */
#undef  DIV0
#define DIV0(N)		({ if (!(N)) break; })

#define DEFFU(FU,DESC)
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)	\
      case OP:								\
	out1 = DEP_##O1; out2 = DEP_##O2;				\
	in1 = DEP_##I1; in2 = DEP_##I2; in3 = DEP_##I3;			\
	EXPR;								\
	break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
      case OP:								\
	/* can speculatively decode a bogus inst */			\
	op = NOP;							\
	out1 = NA; out2 = NA;						\
	in1 = NA; in2 = NA; in3 = NA;					\
	/* no EXPR */							\
	break;
#define CONNECT(OP)
#include "ss.def"
#undef DEFFU
#undef DEFINST
#undef DEFLINK
#undef CONNECT
      default:
	/* can speculatively decode a bogus inst */
	op = NOP;
	out1 = NA; out2 = NA;
	in1 = NA; in2 = NA; in3 = NA;
	/* no EXPR */
      }
      /* operation sets next_PC */

      if (ruu_perfect_prediction)
	{
	  pred_PC = next_PC;
	}

      /* no effective address dep? */
      no_ea_dep = FALSE;

      if (/* skip NOPs */op != NOP)
	{
	  /* for load/stores:
	       idep #0     - store operand (value that is store'ed)
	       idep #1, #2 - eff addr computation inputs (addr of access)

	     resulting RUU/LSQ operation pair:
	       RUU (effective address computation operation):
	         idep #0, #1 - eff addr computation inputs (addr of access)
	       LSQ (memory access operation):
	         idep #0     - operand input (value that is store'd)
		 idep #1     - eff addr computation result (from RUU op)

	     effective address computation is transfered via the reserved
	     name DEP_TMP
	   */

	  /* fill in RUU reservation station */
	  rs = &RUU[RUU_tail];

	  rs->IR = inst;
	  rs->op = op;
	  rs->PC = regs_PC;
	  rs->next_PC = next_PC; rs->pred_PC = pred_PC;
	  rs->in_LSQ = FALSE;
	  rs->ea_comp = FALSE;
	  rs->recover_inst = FALSE;
	  rs->spec_mode = spec_mode;
	  rs->addr = 0;
	  /* rs->tag is already set */
	  rs->seq = ++inst_seq;
	  rs->queued = rs->issued = rs->completed = FALSE;

	  /* split ld/st's into two operations: eff addr comp + mem access */
	  if (SS_OP_FLAGS(op) & F_MEM)
	    {
	      /* convert RUU operation from ld/st to an add (eff addr comp) */
	      rs->op = ADD;
	      rs->ea_comp = TRUE;

	      /* fill in LSQ reservation station */
	      lsq = &LSQ[LSQ_tail];

	      lsq->IR = inst;
	      lsq->op = op;
	      lsq->PC = regs_PC;
	      lsq->next_PC = next_PC; lsq->pred_PC = pred_PC;
	      lsq->in_LSQ = TRUE;
	      lsq->ea_comp = FALSE;
	      lsq->recover_inst = FALSE;
	      lsq->spec_mode = spec_mode;
	      lsq->addr = addr;
	      /* lsq->tag is already set */
	      lsq->seq = ++inst_seq;
	      lsq->queued = lsq->issued = lsq->completed = FALSE;

	      /* link eff addr computation onto operand's output chains */
	      ruu_link_idep(rs, /* idep_ready[] index */0, NA);
	      ruu_link_idep(rs, /* idep_ready[] index */1, in2);
	      ruu_link_idep(rs, /* idep_ready[] index */2, in3);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(rs, /* odep_list[] index */0, DEP_TMP);
	      ruu_install_odep(rs, /* odep_list[] index */1, NA);

	      /* link memory access onto output chain of eff addr operation */
	      ruu_link_idep(lsq,
			    /* idep_ready[] index */STORE_OP_INDEX/* 0 */,
			    in1);
	      ruu_link_idep(lsq,
			    /* idep_ready[] index */STORE_ADDR_INDEX/* 1 */,
			    no_ea_dep ? NA : DEP_TMP);
	      ruu_link_idep(lsq, /* idep_ready[] index */2, NA);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(lsq, /* odep_list[] index */0, out1);
	      ruu_install_odep(lsq, /* odep_list[] index */1, out2);

	      /* install operation in the RUU and LSQ */
	      n_dispatched++;
	      RUU_tail = (RUU_tail + 1) % RUU_size;
	      RUU_num++;
	      LSQ_tail = (LSQ_tail + 1) % LSQ_size;
	      LSQ_num++;

	      if (OPERANDS_READY(rs))
		{
		  /* eff addr computation ready, queue it on ready list */
		  readyq_enqueue(rs);
		}
	      /* issue may continue when the load/store is issued */
	      last_op = RSLINK_INIT(lsq);

	      /* issue stores only, loads are issued by lsq_refresh() */
	      if (((SS_OP_FLAGS(op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
		  && OPERANDS_READY(lsq))
		{
		  /* panic("store immediately ready"); */
		  /* put operation on ready list, ruu_issue() issue it later */
		  readyq_enqueue(lsq);
		}
	    }
	  else /* !(SS_OP_FLAGS(op) & F_MEM) */
	    {
	      /* link onto producing operation */
	      ruu_link_idep(rs, /* idep_ready[] index */0, in1);
	      ruu_link_idep(rs, /* idep_ready[] index */1, in2);
	      ruu_link_idep(rs, /* idep_ready[] index */2, in3);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(rs, /* odep_list[] index */0, out1);
	      ruu_install_odep(rs, /* odep_list[] index */1, out2);

	      /* install operation in the RUU */
	      n_dispatched++;
	      RUU_tail = (RUU_tail + 1) % RUU_size;
	      RUU_num++;

	      /* issue op if all its reg operands are ready (no mem input) */
	      if (OPERANDS_READY(rs))
		{
		  /* put operation on ready list, ruu_issue() issue it later */
		  readyq_enqueue(rs);
		  /* issue may continue */
		  last_op = RSLINK_NULL;
		}
	      else
		{
		  /* could not issue this inst, stall issue until we can */
		  last_op = RSLINK_INIT(rs);
		}
	    }
	}
      else
	{
	  /* indicate results are being created in the decode stage */
	  if (out1)
	    BITMAP_SET(decode_stage_results, BITMAP_SIZE(SS_TOTAL_REGS), out1);
	  if (out2)
	    BITMAP_SET(decode_stage_results, BITMAP_SIZE(SS_TOTAL_REGS), out2);
	  rs = NULL;
	}

      if (!spec_mode)
	{
	  /* one more non-speculative instruction executed */
	  num_insn++;

	  /* update BTB */
	  if (SS_OP_FLAGS(op) & F_CTRL)
	    {
	      if (pred)
		bpred_update(pred, regs_PC, next_PC,
			     /*taken?*/next_PC != (regs_PC +
						   sizeof(SS_INST_TYPE)),
			     /*correct pred? */pred_PC == next_PC); 
	    }

	  /* begin mis-speculation? */
	  if (pred_PC != next_PC)
	    {
	      recover_PC = next_PC;
	      spec_mode = TRUE;
	      rs->recover_inst = TRUE;
	    }
	}

      /* consume instruction */
      fetch_head = (fetch_head+1) & (ruu_decode_width-1);
      fetch_num--;

      /* one more instruction executed, speculative or otherwise */
      num_total_insn++;
    }
}

void
ruu_fetch(void)
{
  int i, lat, done = FALSE;
  SS_INST_TYPE inst;

  for (i=0;
       (i < ruu_decode_width)             /* fetch up to NUMISSUE insts */
       && (fetch_num < ruu_decode_width)  /* fetch until idecode() stage full */
       && !done;
       i++)
    {
      /* fetch instruction */
      fetch_regs_PC = fetch_pred_PC;
      if (ld_text_base <= fetch_regs_PC
          && fetch_regs_PC < (ld_text_base+ld_text_size)
          && !(fetch_regs_PC & (sizeof(SS_INST_TYPE)-1)))
	{
          if (icache)
            {
              lat = cache_access(icache, Read, fetch_regs_PC,
                                 &inst, sizeof(SS_INST_TYPE), cycle, NULL, NULL);
              if (lat != 1)
                {
                  /* I-cache miss */

                  /* may not be the case due to squashes */
                  /* assert(lat == blk_access_latency); */

		  ruu_fetch_issue_delay += lat - 1;
                  break;
                }
              /* else: cache hit */
            }
          else
            mem_access(Read, fetch_regs_PC, &inst, sizeof(SS_INST_TYPE));
	}
      else
	inst = SS_NOP_INST;

      /* have a valid inst, here */

      /* possibly use the BTB target */
      if (pred)
	{
	  fetch_pred_PC = bpred_lookup(pred, fetch_regs_PC, 0);
	  if (!fetch_pred_PC)
	    fetch_pred_PC = fetch_regs_PC + sizeof(SS_INST_TYPE);
	  else
	    {
	      /* discontinuous fetch, terminate */
              done = TRUE;
            }
	}
      else
        {
          /* predicted not taken, continue fetching linearly */
          fetch_pred_PC = fetch_regs_PC + sizeof(SS_INST_TYPE);
        }

      /* commit this instruction to the next stage */
      fetch_data[fetch_tail].IR = inst;
      fetch_data[fetch_tail].regs_PC = fetch_regs_PC;
      fetch_data[fetch_tail].pred_PC = fetch_pred_PC;
      fetch_tail = (fetch_tail + 1) & (ruu_decode_width - 1);
      fetch_num++;
    }
}


/* decode and issue instructions */
void
sim_main(void)
{
  num_insn = 0;
  num_refs = 0;

  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;
    SS_INST_TYPE inst;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    fprintf(stderr, "decoding text segment...");
    for (addr=ld_text_base;
	 addr < (ld_text_base+ld_text_size);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr);
	inst.a = (inst.a & ~0xff) | (unsigned long)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr) = inst;
      }
    fprintf(stderr, "done.\n");
  }

  /* initialize the simulation engine */
  fu_pool = res_create_pool("fu-pool", fu_config, N_ELT(fu_config));
  rslink_init(MAX_RS_LINKS);
  tracer_init();
  fetch_init();
  cv_init();
  eventq_init();
  readyq_init();
  ruu_init();
  lsq_init();

  /* set up program entry state */
  SET_CPC(ld_prog_entry-sizeof(SS_INST_TYPE));
  fetch_regs_PC = regs_PC;
  fetch_pred_PC = ld_prog_entry;

  /* currently RUU issue (from readyq_service(), ruu_writeback(), and
     lsq_refresh() is only limited by the number of functional units */
  for (;;)
    {
      /* RUU/LSQ sanity checks */
      if (RUU_num < LSQ_num)
	panic("RUU_num < LSQ_num");
      if (((RUU_head + RUU_num) % RUU_size) != RUU_tail)
	panic("RUU_head/RUU_tail wedged");
      if (((LSQ_head + LSQ_num) % LSQ_size) != LSQ_tail)
	panic("LSQ_head/LSQ_tail wedged");

      /* commit entries from RUU/LSQ to architected register file */
      ruu_commit();

      /* service function unit release events */
      ruu_release_fu();

      /* ==> may have ready queue entries carried over from previous cycles */

      /* service result completions, also readies dependent operations */
      /* ==> inserts operations into ready queue --> register deps resolved */
      ruu_writeback();

      /* decode and dispatch new operations */
      /* ==> insert ops w/ no deps or all regs ready --> reg deps resolved */
      ruu_dispatch();

      /* try to locate memory operations that are ready to execute */
      /* ==> inserts operations into ready queue --> memory deps resolved */
      lsq_refresh();

      /* issue operations ready to execute from a previous cycle */
      /* <== drains ready queue <-- ready operations commence execution */
      ruu_issue();

      if (!ruu_fetch_issue_delay)
	ruu_fetch();
      else
	ruu_fetch_issue_delay--;

      /* go to next cycle */
      cycle++;
    }
}
